'''This module contains the test cases for the KNN module.'''
import unittest

import numpy as np
import pandas as pd

from chemfusekit.plsda import PLSDASettings, PLSDA, GraphMode
from chemfusekit.lldf import LLDFSettings, LLDF, LLDFDataModel, Table


class TestPLSDA(unittest.TestCase):
    '''Test suite for the PLSDA module.'''

    def test_plsda_settings(self):
        '''Test case against settings errors.'''

        # n_components parameter
        with self.assertRaises(ValueError):
            PLSDASettings(n_components=-3) # Negative value
        with self.assertRaises(ValueError):
            PLSDASettings(n_components=0)  # Zero value
        PLSDASettings(n_components=7)  # Correct value (shouldn't raise anything)

        # output parameter
        with self.assertRaises(TypeError):
            PLSDASettings(output=3)   # Wrong type (not a bool)

        # test_split parameter
        with self.assertRaises(TypeError):
            PLSDASettings(test_split=3)   # Wrong type (not a bool)

        # output and test_split incompatibilities
        with self.assertRaises(Warning):
            PLSDASettings(output=GraphMode.NONE, test_split=True)

    def test_plsda_constructor(self):
        '''Test case against constructor errors.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # settings parameter
        wrong_settings = LLDFDataModel(pd.DataFrame([1]), pd.DataFrame([1]), np.asarray([1]))
        with self.assertRaises(TypeError):
            PLSDA(wrong_settings, lldf.fused_data)  # pass an object of the wrong class as settings

        # fused_data parameter
        knn_settings = PLSDASettings()
        wrong_fused_data = lldf_settings
        with self.assertRaises(TypeError):
            PLSDA(knn_settings, wrong_fused_data)  # pass an object of the wrong class as fused_data
    
    def test_plsda(self):
        '''Integration test case for the training function.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up and run PLSDA (no output)
        plsda_settings = PLSDASettings()
        plsda = PLSDA(plsda_settings, lldf.fused_data)
        plsda.plsda()

        # Set up and run PLSDA with text output
        plsda_settings = PLSDASettings(output=GraphMode.TEXT)
        plsda = PLSDA(plsda_settings, lldf.fused_data)
        plsda.plsda()

        # Set up and run PLSDA with graphical output
        plsda_settings = PLSDASettings(output=GraphMode.GRAPHIC)
        plsda = PLSDA(plsda_settings, lldf.fused_data)
        plsda.plsda()

        # Run with text output and split testing
        plsda_settings = PLSDASettings(output=GraphMode.TEXT, test_split=True)
        plsda = PLSDA(plsda_settings, lldf.fused_data)
        plsda.plsda()

        # Run with graphical output and split testing
        plsda_settings = PLSDASettings(output=GraphMode.GRAPHIC, test_split=True)
        plsda = PLSDA(plsda_settings, lldf.fused_data)
        plsda.plsda() 
    
    def test_prediction(self):
        '''Test case against prediction parameter issues.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up KNN without training it
        plsda_settings = PLSDASettings()
        plsda = PLSDA(plsda_settings, lldf.fused_data)

        # Pick a random sample for prediction
        x_data_sample = lldf.fused_data.x_train.iloc[119] # should be DMMP
        x_data_sample = x_data_sample.iloc[1:].to_frame().transpose()

        # Run prediction with untrained model (should throw exception)
        with self.assertRaises(RuntimeError):
            plsda.predict(x_data_sample)

        # Run training
        plsda.plsda()

        # Run prediction with empty data (should throw exception)
        with self.assertRaises(TypeError):
            plsda.predict(None)

        # Run prediction with trained model and non-null data
        plsda.predict(x_data_sample)



'''This module contains the test cases for the PCA module.'''
import unittest
import copy
from chemfusekit.pca import PCASettings, PCA, GraphMode
from chemfusekit.lldf import LLDFSettings, LLDF, Table
from chemfusekit.lr import LRSettings, LR

class TestPCA(unittest.TestCase):
    '''Test suite for the PCA module.'''

    def test_pca_settings(self):
        '''Test case against settings errors.'''
        # Test for rejection of a negative variance:
        with self.assertRaises(ValueError):
            PCASettings(target_variance=-0.3)

        # Test for rejection of an out-of-bounds confidence level:
        with self.assertRaises(ValueError):
            PCASettings(confidence_level=3.00)

        # Test for rejection of insufficient initial components:
        with self.assertRaises(ValueError):
            PCASettings(initial_components=2)

        # Now try with no mistakes:
        PCASettings(
            target_variance=0.98,
            confidence_level=0.9,
            initial_components=8,
            output=GraphMode.GRAPHIC
        )

    def test_pca_constructor(self):
        '''Test case against constructor parameter issues.'''

        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        lda_settings = PCASettings()

        # First, construct the object with null model:
        with self.assertRaises(TypeError):
            PCA(lda_settings, None)

        # Then, construct the object with null settings:
        with self.assertRaises(TypeError):
            PCA(None, lldf.fused_data)

        # Now, with both null:
        with self.assertRaises(TypeError):
            PCA(None, None)

        # Finally, with proper values:
        PCA(lda_settings, lldf.fused_data)
    
    def test_pca(self):
        '''
        Integration test case to verify that the output does not change based on
        whether the output is set to true or false
        '''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up and execute PCA (graph output)
        pca_settings = PCASettings(output=GraphMode.GRAPHIC)
        pca = PCA(pca_settings, lldf.fused_data)
        pca.pca()

        # Save the results
        result_true_components = copy.deepcopy(pca.components)
        result_true_array_scores = copy.deepcopy(pca.array_scores)

        # Set up and execute PCA (again)
        pca_settings = PCASettings(output=GraphMode.NONE)
        pca = PCA(pca_settings, lldf.fused_data)
        pca.pca()

        # Save the results
        result_false_components = pca.components
        result_false_array_scores = pca.array_scores

        self.assertEqual(result_true_components, result_false_components)
        self.assertEqual(result_true_array_scores, result_false_array_scores)

        # Set up and execute PCA (text output)
        pca_settings = PCASettings(output=GraphMode.GRAPHIC)
        pca = PCA(pca_settings, lldf.fused_data)
        pca.pca()

        # Save the results
        result_true_components = copy.deepcopy(pca.components)
        result_true_array_scores = copy.deepcopy(pca.array_scores)

        # Set up and execute PCA (again)
        pca_settings = PCASettings(output=GraphMode.NONE)
        pca = PCA(pca_settings, lldf.fused_data)
        pca.pca()

        # Save the results
        result_false_components = pca.components
        result_false_array_scores = pca.array_scores

        self.assertEqual(result_true_components, result_false_components)
        self.assertEqual(result_true_array_scores, result_false_array_scores)

    def test_pca_integration_lr(self):
        '''Integration test for PCA+LR'''

        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up PCA and get the rescaled_data property directly
        pca_settings = PCASettings()
        pca = PCA(pca_settings, lldf.fused_data)
        rescaled_data = pca.rescaled_data

        # Set up and execute LR
        lr_settings = LRSettings()
        lr = LR(lr_settings, rescaled_data)
        lr.lr()

    def test_pca_import_export(self):
        '''Test case for the import and export of PCA models.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up PCA
        pca_settings = PCASettings()
        pca = PCA(pca_settings, lldf.fused_data)

        # Try exporting the model before executing pca()
        with self.assertRaises(RuntimeError):
            pca.export_model('pca_model.sklearn')

        # Execute PCA and retry exporting
        pca.pca()
        pca.export_model('pca_model.sklearn')

        # Try creating a new PCA object from the wrong type of file
        with self.assertRaises(ImportError):
            PCA.from_file(pca_settings, 'tests/qepas.xlsx')

        # Create a new PCA object from file
        pca2 = PCA.from_file(pca_settings, 'pca_model.sklearn')

        # Assert the equality of the two models
        self.assertEqual(pca.pca_model.get_params(), pca2.pca_model.get_params())

    def test_pca_reduce(self):
        '''Test case for data dimensionality reduction.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up PCA
        pca_settings = PCASettings()
        pca = PCA(pca_settings, lldf.fused_data)

        # Try rescaling data before training the model
        with self.assertRaises(RuntimeError):
            pca.reduce(lldf.fused_data)

        # Execute PCA and then rescale
        pca.pca()
        reduced_data = pca.reduce(lldf.fused_data)

        # Check that the dimensionality (number of columns) is reduced
        self.assertLess(reduced_data.x_data.shape[1], lldf.fused_data.x_data.shape[1])



'''Support Vector Machine module.'''
from copy import copy
from typing import Optional

import pandas as pd

from sklearn.svm import SVC

from chemfusekit.lldf import LLDFDataModel
from chemfusekit.__utils import GraphMode, run_split_test, print_confusion_matrix
from .__base import BaseSettings, BaseClassifier, BaseDataModel


class SVMSettings(BaseSettings):
    '''Holds the settings for the SVM object.'''
    def __init__(self, kernel: str = 'linear', output: GraphMode = GraphMode.NONE, test_split: bool = False):
        super().__init__(output, test_split)
        if kernel not in ['linear', 'poly', 'gaussian', 'sigmoid']:
            raise ValueError("Invalid type: must be linear, poly, gaussian or sigmoid")
        self.kernel = kernel


class SVM(BaseClassifier):
    '''Class for Support Vector Machine analysis of the data'''
    def __init__(self, settings: SVMSettings, data: BaseDataModel):
        super().__init__(settings, data)

    def svm(self):
        '''Performs Support Vector Machine analysis'''

        # Linear kernel
        if self.settings.kernel == 'linear':
            svm_model = SVC(kernel='linear', probability=True)
        # Polynomial kernel
        elif self.settings.kernel == 'poly':
            svm_model = SVC(kernel='poly', degree=8)
        # Gaussian kernel - radial basis function
        elif self.settings.kernel == 'gaussian':
            svm_model = SVC(kernel='rbf')
        # Sigmoid classifier
        elif self.settings.kernel == 'sigmoid':
            svm_model = SVC(kernel='sigmoid')
        else:
            raise ValueError(f"SVM: this type of kernel does not exist ({self.settings.kernel=})")

        svm_model.fit(self.data.x_data, self.data.y)
        self.model = svm_model

        predictions = svm_model.predict(self.data.x_data)
        print_confusion_matrix(
            self.data.y,
            predictions,
            "Confusion matrix based on the whole data set",
            mode=self.settings.output
        )

        if self.settings.test_split:
            run_split_test(
                x=self.data.x_data,
                y=self.data.y,
                model=SVC(kernel=self.settings.kernel),
                mode=self.settings.output
            )

    def import_model(self, import_path: str):
        model_backup = copy(self.model)
        super().import_model(import_path)
        if not isinstance(self.model, SVC):
            self.model = model_backup
            raise ImportError("The file you tried to import is not an SVC classifier.")
        self.settings.kernel = self.model.kernel



'''Principal Component Analysis Module'''
from copy import copy
from functools import cached_property
from typing import Optional

import joblib
import numpy as np
import pandas as pd
import plotly.express as px
import matplotlib.pyplot as plt

from sklearn.decomposition import PCA as PC

import scipy.stats

from chemfusekit.__utils import print_table, GraphMode
from .__base import BaseDataModel


class PCADataModel(BaseDataModel):
    '''Data model for the PCA outputs.'''
    def __init__(self, x_data: pd.DataFrame, x_train: pd.DataFrame, y: np.ndarray, array_scores: np.ndarray,
                 components: int):
        super().__init__(x_data, x_train, y)
        self.array_scores = array_scores
        self.components = components


class PCASettings:
    '''Holds the settings for the PCA object.'''
    def __init__(self, target_variance: float = 0.95,
                 confidence_level: float = 0.05,
                 initial_components: int = 10, output: GraphMode = GraphMode.NONE):
        if target_variance < 0:
            raise ValueError("Target variance should be positive or null.")
        if confidence_level < 0 or confidence_level > 1:
            raise ValueError("Confidence level should be between 0 and 1.")
        if initial_components < 3:
            raise ValueError("Initial components should be at least 3.")
        self.target_variance = target_variance
        self.confidence_level = confidence_level
        self.initial_components = initial_components
        self.output = output


class PCA:
    '''A class to store the data, methods and artifacts for Principal Component Analysis'''
    def __init__(self, settings: PCASettings, data: BaseDataModel):
        self.data = data
        self.components = 0
        self.pca_model: Optional[PC] = None
        self.settings = settings
        self.array_scores: Optional[np.ndarray] = None

    def pca(self):
        '''Performs Principal Component Analysis.'''

        # Read from the data fusion object
        x_data = self.data.x_data

        # Run PCA producing the reduced variable Xreg and select the first 10 components
        pca = PC(self.settings.initial_components)
        pca.fit(x_data)

        # Define the class vector (discrete/categorical variable)
        # y_dataframe = pd.DataFrame(self.fused_data.y, columns=['Substance'])
        # classes = y_dataframe.astype('category') (a cosa serve?)
        out_sum = np.cumsum(pca.explained_variance_ratio_)

        # Autoselect the number of components
        for i, x in enumerate(out_sum):
            if x >= self.settings.target_variance:
                self.components = i
                break
        self.components = max(self.components, 3)

        compsexpv = [[(i+1), pca.explained_variance_ratio_[i]] for i in np.arange(pca.n_components_)]
        comps, expv = zip(*compsexpv)
        print_table(
            ["Components", "Explained Variance"],
            [comps, expv],
            "Proportion of Variance Explained",
            mode=self.settings.output
        )

        if self.settings.output is GraphMode.GRAPHIC:
            # PCA scree plot
            pc_values = np.arange(pca.n_components_) + 1
            plt.plot(pc_values, pca.explained_variance_ratio_, 'ro-', linewidth=2)
            plt.title('Scree Plot')
            plt.xlabel('Principal Component')
            plt.ylabel('Proportion of Variance Explained')
            plt.show()

        compsexpv = [[(i+1), out_sum[i]] for i in np.arange(pca.n_components_)]
        comps, expv = zip(*compsexpv)
        print_table(
            ["Components", "Cumulative Explained Variance"],
            [comps, expv],
            "Cumulative Proportion of Variance Explained",
            mode=self.settings.output
        )

        if self.settings.output is GraphMode.GRAPHIC:
            # Cumulative explained variance ratio
            plt.plot(pc_values, out_sum, 'ro-', linewidth=2)
            plt.title('Scree Plot (cumulative)')
            plt.xlabel('Principal Component')
            plt.ylabel('Cumulative Proportional Variance Explained')
            plt.show()

        # Run PCA producing the pca_model with a proper number of components
        pca = PC(n_components=self.components)
        self.pca_model = pca
        self.pca_model.fit(x_data)

    def pca_stats(self):
        '''Produces PCA-related statistics.'''
        x_data = self.data.x_data
        x_train = self.data.x_train

        # Prepare the Scores dataframe (and concatenate the original 'Region' variable)
        pc_cols = [f"PC{i+1}" for i in range(self.components)]
        scores = pd.DataFrame(data=self.pca_model.fit_transform(x_data), columns=pc_cols)
        scores.index = x_data.index
        scores = pd.concat([scores, x_train.Substance], axis = 1)

        print_table(
            pc_cols + ['Substance'],
            [scores.iloc[:,i] for i in range(scores.shape[1])],
            "PCA scores for each component",
            self.settings.output
        )

        # Prepare the loadings dataframe
        loadings = pd.DataFrame(
            self.pca_model.components_.T,
            columns=pc_cols,
            index=x_data.columns
        )
        loadings["Attributes"] = loadings.index

        print_table(
            pc_cols + ['Retention Time'],
            [loadings.iloc[:,i] for i in range(loadings.shape[1])],
            "PCA Loadings",
            self.settings.output
        )

        if self.settings.output is GraphMode.GRAPHIC:
            # View the scores plot using plotly library
            fig = px.scatter(
                scores,
                x="PC1",
                y="PC2",
                color="Substance",
                hover_data=['Substance'],
                hover_name=x_data.index
            )
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_layout(
                height=600,
                width=800,
                title_text='PCA Scores Plot colored by Substance')
            fig.show()

            # Plot 3D scores
            fig = px.scatter_3d(
                scores,
                x='PC1',
                y='PC2',
                z='PC3',
                color='Substance',
                hover_data=['Substance'],
                hover_name=x_data.index
            )
            fig.show()

        # Get PCA scores
        t = scores.iloc[:,0:self.components]
        # Get PCA loadings
        p = loadings.iloc[:,0:self.components]
        # Calculate error array
        err = x_data - np.dot(t,p.T)
        # Calculate Q-residuals (sum over the rows of the error array)
        q= np.sum(err**2, axis=1)
        # Calculate Hotelling's T-squared (note that data are normalised by default)
        tsq = np.sum((t/np.std(t, axis=0))**2, axis=1)

        def mean_confidence_interval(data, confidence=self.settings.confidence_level):
            a = 1.0 * np.array(data)
            n = len(a)
            m, se = np.mean(a), scipy.stats.sem(a)
            h = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
            return m, m-h, m+h

        tsq_conf = (mean_confidence_interval(
            tsq.values,
            confidence=self.settings.confidence_level)
        )[2]
        q_conf = (mean_confidence_interval(
            q.values,
            confidence=self.settings.confidence_level)
        )[2]

        # Create a dataframe using only T2 and Q-residuals
        hot_q_data = pd.DataFrame(
            {'T2': tsq, 'Qres': q, 'Substance': x_train.Substance},
            index = x_data.index
        )

        if self.settings.output is GraphMode.GRAPHIC:
            # Plot the Hotelling T2 vs Q-residuals plot
            fig = px.scatter(
                hot_q_data,
                x="T2",
                y="Qres",
                hover_data={'Sample': (hot_q_data.index)},
                color = "Substance"
            )
            fig.add_hline(y=abs(q_conf),line_dash="dot", line_color='Red')
            fig.add_vline(x=tsq_conf,line_dash="dot", line_color='Red')
            fig.update_traces(textposition='top center')
            fig.update_layout(
                height=600,
                width=800,
                title_text="Hotelling's T2 vs Q-residuals")
            fig.show()

        # Normalize Q-residuals and Hotelling's T-squared
        normalized_q = q / np.max(q)
        normalized_tsq = tsq / np.max(tsq)

        # Create a DataFrame with normalized values
        normalized_hot_q_data = {
            'T2': normalized_tsq,
            'Qres': normalized_q,
            'Substance': x_train.Substance
        }
        normalized_hot_q_data = pd.DataFrame(normalized_hot_q_data, index=x_data.index)

        if self.settings.output is GraphMode.GRAPHIC:
            # Plot the normalized Hotelling T2 vs Q-residuals plot
            fig_normalized = px.scatter(
                normalized_hot_q_data,
                x="T2",
                y="Qres",
                hover_data={'Sample': (normalized_hot_q_data.index)},
                color="Substance"
            )
            fig_normalized.add_hline(y=abs(q_conf / np.max(q)), line_dash="dot", line_color='Red')
            fig_normalized.add_vline(x=tsq_conf / np.max(tsq), line_dash="dot", line_color='Red')
            fig_normalized.update_traces(textposition='top center')
            fig_normalized.update_layout(
                height=600,
                width=800,
                title_text="Normalized Hotelling's T2 vs Q-residuals"
            )
            fig_normalized.show()

        # Assuming 'scores' is your DataFrame with the 'class' column
        # Drop the 'class' column before converting to NumPy array
        array_scores = scores.drop('Substance', axis=1).values

        print_table(
            pc_cols,
            [array_scores[:,i] for i in range(array_scores.shape[1])],
            "Array without 'Substance' column",
            self.settings.output
        )

        self.array_scores = array_scores

    def export_data(self) -> PCADataModel:
        '''Export data artifacts.'''
        if self.pca_model is None or self.array_scores is None:
            raise RuntimeError("Run both pca() and pca_stats() methods before exporting data!")

        return PCADataModel(
            self.data.x_data,
            self.data.x_train,
            self.data.y,
            self.array_scores,
            self.components
        )

    @cached_property
    def rescaled_data(self) -> PCADataModel:
        if self.array_scores is None:
            settings_backup = copy(self.settings)
            self.settings.output = GraphMode.NONE
            if self.pca_model is None:
                self.pca()
            self.pca_stats()

        x_data = pd.DataFrame(self.pca_model.transform(self.data.x_data))
        y_dataframe = pd.DataFrame(self.data.y, columns=['Substance'])
        x_train = pd.concat(
            [y_dataframe, x_data],
            axis=1
        )

        return PCADataModel(
            x_data,
            x_train,
            self.data.y,
            self.array_scores,
            self.components
        )

    @classmethod
    def from_file(cls, settings: PCASettings, model_path: str):
        '''Creates a PCA instance from a file containing its sklearn core model.'''
        try:
            model = joblib.load(model_path)
        except Exception as exc:
            raise ImportError("The file you tried importing is not a valid Python object!") from exc
        if not isinstance(model, PC):
            raise ImportError("The file you tried importing is not a sklearn PCA model!")
        data = BaseDataModel(
            pd.DataFrame(),
            pd.DataFrame(),
            np.asarray(pd.DataFrame)
        )
        class_instance = PCA(settings, data)
        class_instance.pca_model = model
        return class_instance

    def export_model(self, export_path: str):
        '''Exports the underlying sklearn PCA model to a file.'''
        if self.pca_model is not None:
            joblib.dump(self.pca_model, export_path)
        else:
            raise RuntimeError("You haven't trained the model yet! You cannot export it now.")

    def reduce(self, data: BaseDataModel) -> BaseDataModel:
        '''Reduces dimensionality of data.'''
        if self.pca_model is None:
            raise RuntimeError(
                "The PCA model hasn't been trained yet! You cannot use it to reduce data dimensionality."
            )
        x_data = pd.DataFrame(self.pca_model.transform(data.x_data))
        y_dataframe = pd.DataFrame(data.y)
        x_train = pd.concat(
            [y_dataframe, x_data],
            axis=1
        )
        return BaseDataModel(
            x_data=x_data,
            x_train=x_train,
            y=data.y
        )



'''k-Nearest Neighbors Analysis module'''
from copy import copy
from typing import Optional
from beartype.typing import Callable

from sklearn.neighbors import KNeighborsClassifier

import pandas as pd

from chemfusekit.lldf import LLDFDataModel
from chemfusekit.__utils import run_split_test, print_confusion_matrix, print_table, GraphMode
from .__base import BaseSettings, BaseClassifier


class KNNSettings(BaseSettings):
    '''Holds the settings for the kNN object.'''
    def __init__(self, n_neighbors: int = 15, metric: str | Callable = 'euclidean', weights: str | Callable = 'uniform',
                 algorithm: str = 'auto', output: GraphMode = GraphMode.NONE, test_split: bool = False):

        super().__init__(output, test_split)

        if n_neighbors < 1:
            raise ValueError("Invalid n_neighbors number: should be a positive integer.")
        if metric not in ['minkwoski', 'precomputed', 'euclidean'] and not callable(metric):
            raise ValueError(
                "Invalid metric: should be 'minkwoski', 'precomputed', 'euclidean' or a callable."
            )
        if weights not in ['uniform', 'distance'] and not callable(weights):
            raise ValueError("Invalid weight: should be 'uniform', 'distance' or a callable")
        if algorithm not in ['auto', 'ball_tree', 'kd_tree', 'brute']:
            raise  ValueError(
                "Invalid algorithm: should be 'auto', 'ball_tree', 'kd_tree' or 'brute'."
            )
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.weights = weights
        self.algorithm = algorithm


class KNN(BaseClassifier):
    '''Class to store the data, methods and artifacts for k-Nearest Neighbors Analysis'''
    def __init__(self, settings: KNNSettings, fused_data: LLDFDataModel):
        super().__init__(settings, fused_data)

    def knn(self):
        '''Performs k-Nearest Neighbors Analysis'''
        # Prepare and train the kNN model
        knn = KNeighborsClassifier(
            n_neighbors=self.settings.n_neighbors,
            metric=self.settings.metric,
            weights=self.settings.weights,
            algorithm=self.settings.algorithm
        )
        knn.fit(self.data.x_data, self.data.y)

        # Save the trained model
        self.model = knn

        # View the prediction on the test data
        y_pred = knn.predict(self.data.x_data)
        print_table(
            ["Predictions"],
            y_pred.reshape(1,len(y_pred)),
            "Data and predictions",
            self.settings.output
        )

        print_confusion_matrix(
            self.data.y,
            y_pred,
            "Confusion Matrix based on the whole data set",
            self.settings.output
        )

        if self.settings.test_split and self.settings.output:
            knn_split = KNeighborsClassifier(
                n_neighbors=self.settings.n_neighbors,
                metric=self.settings.metric,
                weights=self.settings.weights,
                algorithm=self.settings.algorithm
            )
            run_split_test(self.data.x_data, self.data.y, knn_split)

    def import_model(self, import_path: str):
        model_backup = copy(self.model)
        super().import_model(import_path)
        if not isinstance(self.model, KNeighborsClassifier):
            self.model = model_backup
            raise ImportError("The file you tried to import is not a KNeighborsClassifier.")
        self.settings.n_neighbors = self.model.n_neighbors
        self.settings.metric = self.model.metric
        self.settings.weights = self.model.weights
        self.settings.algorithm = self.model.algorithm



'''Logistic Regression Module'''
from copy import copy
from typing import Optional

import numpy as np
import pandas as pd

from sklearn.linear_model import LogisticRegression

from chemfusekit.__utils import run_split_test, print_confusion_matrix, print_table, GraphMode
from .__base import BaseSettings, BaseDataModel, BaseClassifier
from .pca import PCADataModel


class LRSettings(BaseSettings):
    '''Holds the settings for the LR object.'''
    def __init__(self, algorithm: str = 'liblinear', output: GraphMode = GraphMode.NONE, test_split: bool = False):
        super().__init__(output, test_split)
        if algorithm not in [
            'lbfgs',
            'liblinear',
            'newton-cg',
            'newton-cholesky',
            'sag',
            'saga'
        ]:
            raise ValueError(f"{algorithm}: this algorithm does not exist.")
        self.algorithm = algorithm


class LR(BaseClassifier):
    '''Class to store the data, methods and artifacts for Logistic Regression'''
    def __init__(self, settings: LRSettings, data: BaseDataModel):
        super().__init__(settings, data)
        if isinstance(data, PCADataModel):
            self.array_scores = data.array_scores
        else:
            self.array_scores = data.x_train.drop('Substance', axis=1).values

    def lr(self):
        '''Performs Logistic Regression'''

        # Let's build our model on the training set
        model = LogisticRegression(
            solver=self.settings.algorithm,
            random_state=0,
            class_weight='balanced',
            max_iter=10000
        ).fit(self.array_scores, self.data.y)

        # See the classes the model used
        classes = np.unique(self.data.y)
        if len(classes) == 2:   # Binary classifier
            classes = [" / ".join(classes)]
            classes = np.asarray(classes)
            classes = classes.reshape((1, 1))
        else:   # Non-binary classifier
            classes = classes.reshape((1, len(classes)))
        coefficients = model.coef_.transpose()
        intercepts = model.intercept_.reshape((1, len(model.intercept_)))

        print_table(
            ["Class"] + [f"Coefficient {i+1}" for i in range(model.coef_.shape[1])] + ["Intercept (bias)"],
            np.concatenate((classes, coefficients, intercepts)),
            "LR Coefficients",
            self.settings.output
        )

        '''
        Evaluate the model: each sample has a probability of belonging to Positive
        or Negative outcome. Class 0 is Negative, class 1 is Positive.  If the value
        of the first column (probability of being Negative) is higher than 0.5, we
        have a Negative sample. Otherwise, it will be Positive
        '''

        probabilities = model.predict_proba(self.array_scores)
        predictions = model.predict(self.array_scores)

        # This tells us the accuracy of our model in calibration
        scores = model.score(self.array_scores, self.data.y)

        # Save the trained model
        self.model = model

        sample_column = self.data.y.reshape((1, self.data.y.shape[0]))
        pred_column = predictions.reshape((1, predictions.shape[0]))
        prob_column = probabilities.transpose()

        classes = np.unique(self.data.y)

        print_table(
            np.concatenate((np.asarray(["Real sample", "Prediction"]), classes)),
            np.concatenate((
                sample_column,
                pred_column,
                prob_column
            )),
            f"LR Predictions with class probabilities (overall score: {scores})",
            self.settings.output
        )

        print_confusion_matrix(
            self.data.y,
            predictions,
            "Confusion Matrix based on whole data set",
            self.settings.output
        )

        if self.settings.test_split:
            split_model = LogisticRegression(
                solver=self.settings.algorithm,
                random_state=0,
                class_weight='balanced',
                max_iter=10000
            )

            run_split_test(
                self.array_scores,
                self.data.y,
                split_model,
                extended=True,
                mode=self.settings.output
            )

    def predict(self, x_sample: pd.DataFrame):
        '''Performs LR prediction once the model is trained.'''
        prediction = super().predict(x_sample)
        probabilities = self.model.predict_proba(x_sample)

        classes = self.model.classes_.reshape((self.model.classes_.shape[0], ))
        probabilities = probabilities.reshape((self.model.classes_.shape[0], ))
        print_table(
            classes,
            probabilities,
            f"Predicted class: {prediction}. Scores:",
            mode=self.settings.output
        )

        return prediction

    def import_model(self, import_path: str):
        model_backup = copy(self.model)
        super().import_model(import_path)
        if not isinstance(self.model, LogisticRegression):
            self.model = model_backup
            raise ImportError("The file you tried to import is not a LogisticRegression classifier.")
        self.settings.algorithm = self.model.solver



'''A base class for all classifiers.'''

import pandas as pd
import numpy as np
import joblib

from sklearn.base import BaseEstimator

from .__utils import GraphMode


class BaseDataModel:
    '''Models the output data from data-outputting operations'''

    def __init__(self, x_data: pd.DataFrame, x_train: pd.DataFrame, y: np.ndarray):
        self.x_data = x_data
        self.x_train = x_train
        self.y = y

    @classmethod
    def load_from_file(cls, import_path: str, sheet_name: str = 'Sheet1', class_column: str = 'Substance',
                       index_column: str | None = None):
        try:
            # Autodetect the format based on the file extension
            if import_path.endswith('.xlsx'):
                table_data = pd.read_excel(
                    import_path,
                    sheet_name=sheet_name,
                    index_col=0,
                    header=0
                )
            elif import_path.endswith('.csv'):
                table_data = pd.read_csv(
                    import_path,
                    index_col=0,
                    header=0
                )
            elif import_path.endswith('.json'):
                table_data = pd.read_json(
                    import_path,
                    orient='table'  # or other orientations based on your json format
                )
            else:
                raise ValueError(f"Unsupported file format: {import_path}")
        except Exception as exc:
            raise FileNotFoundError("Error opening the selected files.") from exc

        if index_column is not None:
            x = table_data.drop(index_column, axis=1)
        else:
            x = table_data

        # It is necessary to convert the column names as string to select them
        x.columns = x.columns.astype(str)  # to make the colnames as text

        # Reset the index of the dataframe
        # x = x.reset_index(drop=True)

        y = table_data.loc[:, class_column].values
        x_train = x
        x_data = x_train.drop(class_column, axis=1)

        return cls(x_data, x_train, y)

    def export_to_file(self, export_path: str, sheet_name: str = 'Sheet1'):
        # Determine the file format based on the file extension
        if export_path.endswith('.xlsx'):
            try:
                self.x_train.to_excel(excel_writer=export_path, sheet_name=sheet_name)
            except Exception as exc:
                raise RuntimeError("Could not export data to the selected path.") from exc
        elif export_path.endswith('.csv'):
            try:
                self.x_train.to_csv(export_path, index=False)
            except Exception as exc:
                raise RuntimeError("Could not export data to the selected path.") from exc
        elif export_path.endswith('.json'):
            try:
                self.x_train.to_json(export_path, orient='table')  # or other orientations based on your needs
            except Exception as exc:
                raise RuntimeError("Could not export data to the selected path.") from exc
        else:
            raise ValueError(f"Unsupported file format: {export_path}")

    def __getitem__(self, index):
        '''Get an item with array-style indexing'''
        return pd.DataFrame(self.x_data.iloc[index, :]).transpose()


class BaseSettings:
    '''Holds the settings for the BaseClassifier object.'''

    def __init__(self, output: GraphMode = GraphMode.NONE, test_split: bool = False):
        if test_split is True and output is GraphMode.NONE:
            raise Warning(
                "You selected test_split but it won't run because you disabled the output."
            )
        self.output = output
        self.test_split = test_split


class BaseClassifier:
    '''Parent class for all classifiers, containing basic shared utilities.'''

    def __init__(self, settings: BaseSettings, data: BaseDataModel):
        self.settings = settings
        self.data = data
        self.model: BaseEstimator | None = None

    def import_model(self, import_path: str):
        model = joblib.load(import_path)
        if not isinstance(model, BaseEstimator):
            raise ImportError("The file you tried importing is not a sklearn model!")
        self.model = model

    def export_model(self, export_path: str):
        if self.model is not None:
            joblib.dump(self.model, export_path)
        else:
            raise RuntimeError("You haven't trained the model yet! You cannot export it now.")

    @classmethod
    def from_file(cls, settings, model_path):
        '''Creates a classifier instance from file'''
        x_data = pd.DataFrame()
        y_dataframe = pd.DataFrame(columns=['Substance'])
        x_train = pd.concat([y_dataframe, x_data], axis=1)
        y = np.asarray(y_dataframe)
        data = BaseDataModel(
            x_data=x_data,
            x_train=x_train,
            y=y
        )
        class_instance = cls(settings, data)
        class_instance.import_model(model_path)
        return class_instance

    def predict(self, x_data: pd.DataFrame):
        '''Performs prediction once the model is trained.'''
        if x_data is None:
            raise TypeError(f"X data for {self.__class__.__name__} prediction must be non-empty.")
        if self.model is None:
            raise RuntimeError(f"The {self.__class__.__name__} model is not trained yet!")

        y_pred = self.model.predict(x_data)
        return y_pred



'''This module contains the test cases for the LR module.'''
import unittest

import numpy as np
import pandas as pd

from chemfusekit.lldf import LLDFSettings, LLDF, Table
from chemfusekit.pca import PCASettings, PCA, PCADataModel
from chemfusekit.lr import LRSettings, LR, GraphMode

class TestLR(unittest.TestCase):
    '''Test suite for the LR module.'''

    def test_lr_settings(self):
        '''Test case against settings input errors.'''
        # Should raise an exception when the algorithm is not available
        with self.assertRaises(ValueError):
            LRSettings(
                algorithm='unknown',
                output=GraphMode.NONE,
                test_split=False
            )

        # Should raise an exception when any of the inputs is a null value
        with self.assertRaises(TypeError):
            LRSettings(
                algorithm=None,
                output=GraphMode.NONE,
                test_split=True
            )
        with self.assertRaises(TypeError):
            LRSettings(
                algorithm='liblinear',
                output=None,
                test_split=False
            ) 
        with self.assertRaises(TypeError):
            LRSettings(
                algorithm='liblinear',
                output=GraphMode.NONE,
                test_split=None
            )
        
        # Check if split tests with no output cause warnings:
        with self.assertRaises(Warning):
            LRSettings(output=GraphMode.NONE, test_split=True)

        # Should not raise any exception when the input is correct
        LRSettings(
            algorithm='liblinear',
            output=GraphMode.TEXT
        )

    def test_lr_constructor(self):
        '''Test case against constructor input errors.'''
        data_model = PCADataModel(
            x_data=pd.DataFrame(),
            x_train=pd.DataFrame(),
            y=np.asarray([7.854, 1.543, 93.55]),
            array_scores=np.asarray([21.0, 3.44, 7.65]),
            components=5
        )
        # Should raise an exception when the inputs are null
        with self.assertRaises(TypeError):
            LR(
                settings=None,
                data=data_model
            )
        with self.assertRaises(TypeError):
            LR(
                settings=LRSettings(),
                data=None
            )

        # Should not raise any exception when all inputs are valid
        LR(
            settings=LRSettings(),
            data=data_model
        )

    def test_lr(self):
        '''Integration test case for LR training.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        pca_settings = PCASettings()
        pca = PCA(pca_settings, lldf.fused_data)
        pca.pca()
        pca.pca_stats()

        pca_data = pca.export_data()

        # With no output
        lr_settings = LRSettings()
        lr = LR(lr_settings, pca_data)
        lr.lr()

        # With text output
        lr_settings = LRSettings(output=GraphMode.TEXT)
        lr = LR(lr_settings, pca_data)
        lr.lr()

        # With graph output
        # With text output
        lr_settings = LRSettings(output=GraphMode.GRAPHIC)
        lr = LR(lr_settings, pca_data)
        lr.lr()

        # With text output and split tests
        lr_settings = LRSettings(output=GraphMode.TEXT, test_split=True)
        lr = LR(lr_settings, pca_data)
        lr.lr()

        # With graph output and split tests
        lr_settings = LRSettings(output=GraphMode.GRAPHIC, test_split=True)
        lr = LR(lr_settings, pca_data)
        lr.lr()

        # A final test with just the LLDF data:
        lr_settings = LRSettings()
        lr = LR(lr_settings, lldf.fused_data)
        lr.lr()

    def test_lr_predict(self):
        '''Test case against prediction input errors.'''
        # Set up the model
        lr_settings = LRSettings()
        pca_data = PCADataModel(
            x_data=pd.DataFrame(),
            x_train=pd.DataFrame(),
            y=np.asarray([7.02, 8.11]),
            array_scores=np.asarray([43.1, 0.06]),
            components=4
        )
        lr = LR(lr_settings, pca_data)

        # Shold raise an exception when started with no data
        with self.assertRaises(TypeError):
            lr.predict(None)

        # Should raise an exception when started with no training
        with self.assertRaises(RuntimeError):
            lr.predict(pd.DataFrame([4.03, 3.14]))



'''This module contains the test cases for the KNN module.'''
import unittest

import pandas as pd
import numpy as np

from chemfusekit.knn import KNNSettings, KNN, GraphMode
from chemfusekit.lldf import LLDFSettings, LLDF, LLDFDataModel, Table


class TestKNN(unittest.TestCase):
    '''Test suite for the KNN module.'''

    def test_knn_settings(self):
        '''Test case against settings errors.'''

        # n_neighbors parameter
        with self.assertRaises(ValueError):
            KNNSettings(n_neighbors=-3) # Negative value
        with self.assertRaises(ValueError):
            KNNSettings(n_neighbors=0)  # Zero value
        KNNSettings(n_neighbors=7)  # Correct value (shouldn't raise anything)

        # metric parameter
        with self.assertRaises(ValueError):
            KNNSettings(metric='invalid attribute') # Non-existent technique
        with self.assertRaises(TypeError):
            value = 3
            KNNSettings(metric=value)   # Non-callable

        for x in ['minkwoski', 'precomputed', 'euclidean']:
            KNNSettings(metric=x)   # Correct values (shouldn't raise anything)
        x = lambda a: a + 10
        KNNSettings(metric=x)   # Pass a callable (shoulnd't raise anything)

        # weights parameter
        with self.assertRaises(ValueError):
            KNNSettings(weights='invalid attribute') # Non-existent technique
        with self.assertRaises(TypeError):
            value = 3
            KNNSettings(weights=value)   # Non-callable

        for x in ['uniform', 'distance']: # Correct values (shouldn't raise anything)
            KNNSettings(weights=x)
        x = lambda a: a + 10
        KNNSettings(weights=x)   # Pass a callable (shoulnd't raise anything)


        # algorithm parameter
        with self.assertRaises(ValueError):
            KNNSettings(algorithm='invalid attribute') # Non-existent technique
        for x in ['auto', 'ball_tree', 'kd_tree', 'brute']:
            KNNSettings(algorithm=x)    # Correct values (shouldn't raise anything)

        # output parameter
        with self.assertRaises(TypeError):
            KNNSettings(output=3)   # Wrong type (not a GraphMode enum)

        # test_split parameter
        with self.assertRaises(TypeError):
            KNNSettings(test_split=3)   # Wrong type (not a bool)

        # output and test_split incompatibilities
        with self.assertRaises(Warning):
            KNNSettings(output=GraphMode.NONE, test_split=True)

    def test_knn_constructor(self):
        '''Test case against constructor errors.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # settings parameter
        wrong_settings = LLDFDataModel(pd.DataFrame([1]), pd.DataFrame([1]), np.asarray([1]))
        with self.assertRaises(TypeError):
            KNN(wrong_settings, lldf.fused_data)  # pass an object of the wrong class as settings

        # fused_data parameter
        knn_settings = KNNSettings()
        wrong_fused_data = lldf_settings
        with self.assertRaises(TypeError):
            KNN(knn_settings, wrong_fused_data)  # pass an object of the wrong class as fused_data

    def test_knn(self):
        '''Integration test case for the training function.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up and run KNN (no output)
        knn_settings = KNNSettings()
        knn = KNN(knn_settings, lldf.fused_data)
        knn.knn()

        # With graph output
        knn_settings = KNNSettings(output=GraphMode.GRAPHIC)
        knn = KNN(knn_settings, lldf.fused_data)
        knn.knn()

        # With text output
        knn_settings = KNNSettings(output=GraphMode.TEXT)
        knn = KNN(knn_settings, lldf.fused_data)
        knn.knn()

    def test_prediction(self):
        '''Test case against prediction parameter issues.'''
        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Set up KNN without training it
        knn_settings = KNNSettings()
        knn = KNN(knn_settings, lldf.fused_data)

        # Pick a random sample for prediction
        x_data_sample = lldf.fused_data.x_train.iloc[119] # should be DMMP
        x_data_sample = x_data_sample.iloc[1:].to_frame().transpose()

        # Run prediction with untrained model (should throw exception)
        with self.assertRaises(RuntimeError):
            knn.predict(x_data_sample)

        # Run training
        knn.knn()

        # Run prediction with empty data (should throw exception)
        with self.assertRaises(TypeError):
            knn.predict(None)

        # Run prediction with trained model and non-null data
        knn.predict(x_data_sample)



'''Utilities module: functions that are shared between different classes'''
from sklearn.cross_decomposition import PLSRegression as PLSR
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix, classification_report

import numpy as np
import pandas as pd

import plotly.express as px
import plotly.graph_objects as go
import seaborn as sns
import matplotlib.pyplot as plt

from tabulate import tabulate

from enum import Enum, auto

class GraphMode(Enum):
    TEXT = auto()
    GRAPHIC = auto()
    NONE = auto()

def graph_output(scores, model, name: str, mode: GraphMode = GraphMode.GRAPHIC):
    '''A reusable graphing function.'''
    # Return early if output is disabled
    if mode is GraphMode.NONE:
        return

    # Text-mode output
    if mode is GraphMode.TEXT:
        print(f"{name} scores:\n {scores}")

        print(f"""
            explained variance ratio with {name}:
            {model.explained_variance_ratio_}
        """)

        # Display the explained variance ratio
        print("Explained Variance Ratio:", model.explained_variance_ratio_)
    
    if mode is GraphMode.GRAPHIC:
        # Scores table
        print_table(
            scores.columns,
            [scores.iloc[:, i] for i in range(scores.shape[1])],
            f"{name} Scores"
        )

        #Scores plot
        fig = px.scatter(scores, x="LV1", y="LV2", color="Substance", hover_data=['Substance'])
        fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
        fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
        fig.update_layout(
            height=600,
            width=800,
            title_text='Scores Plot')
        fig.show()

        # Plot 3D scores
        fig = px.scatter_3d(scores, x='LV1', y='LV2', z='LV3',
                            color='Substance', hover_data=['Substance'],
                            hover_name=scores.index
        )
        fig.update_layout(title_text=f"3D colored by Substance for {name}")
        fig.show()


def print_table(header_values, cell_values, title: str, mode: GraphMode = GraphMode.GRAPHIC):
    '''Multimodal table printing utility.'''
    # Return early if output is disabled
    if mode is GraphMode.NONE:
        return
    # Graphical table printing
    elif mode is GraphMode.GRAPHIC:
        fig = go.Figure(data=[go.Table(
        header=dict(values=header_values,
                    fill_color='paleturquoise',
                    align='left'),
        cells=dict(values=cell_values,
                fill_color='lavender',
                align='left'))
        ])
        fig.update_layout(title=title)
        fig.show()

    # Text-based table printing
    else:
        # Print the title
        print(f"\n\n{title}:")

        # Print the table using tabulate
        print(tabulate(np.transpose(cell_values), headers=header_values, tablefmt="rounded_outline"))


def run_split_test(x, y, model, extended=False, mode: GraphMode = GraphMode.GRAPHIC):
    '''A function to run split tests on trained models.'''

    # Return early if there's nothing to print
    if mode is GraphMode.NONE:
        return

    x_train, x_test, y_train, y_test = train_test_split(
        x,
        y,
        test_size=0.3,
        random_state=42,
        shuffle=True,
        stratify=y
    )

    model.fit(x_train, y_train)
    y_pred = model.predict(x_train)

    # TODO: add something to print _x_scores multimodally

    if extended:
        probabilities = model.predict_proba(x_train)
        score = model.score(x_train, y_train)
        predictions = model.predict(x_train)

        # See the classes the model used
        classes = model.classes_
        if isinstance(model, LogisticRegression) and len(classes) == 2:     # Binary classifier
            classes = [" / ".join(classes)]
            classes = np.asarray(classes)
            classes = classes.reshape((1, 1))
        else:   # All other cases
            classes = classes.reshape((1, len(classes)))

        coefficients = model.coef_.transpose()
        intercepts = model.intercept_.reshape((1, len(model.intercept_)))
        print_table(
            ["Class"] + [f"Coefficient {i + 1}" for i in range(model.coef_.shape[1])] + ["Intercept (bias)"],
            np.concatenate((classes, coefficients, intercepts)),
            "LR Coefficients + Intercept (split test, training set)",
            mode
        )

        # Training predictions table
        sample_column = y_train.reshape((1, y_train.shape[0]))
        pred_column = predictions.reshape((1, predictions.shape[0]))
        prob_column = probabilities.transpose()

        classes = np.unique(y_train)

        print_table(
            np.concatenate((np.asarray(["True", "Prediction"]), classes)),
            np.concatenate((
                sample_column,
                pred_column,
                prob_column
            )),
            f"LR Predictions with class probabilities (split test, evaluation set, overall score: {score})",
            mode
        )

        print_confusion_matrix(
            y1=y_train,
            y2=predictions,
            title="Confusion matrix based on training set",
            mode=mode
        )

        # Test predictions set table
        probabilities = model.predict_proba(x_test)
        score = model.score(x_test, y_test)
        predictions = model.predict(x_test)
        pred_column = predictions.reshape((1, predictions.shape[0]))
        prob_column = probabilities.transpose()

        print_table(
            np.concatenate((np.asarray(["Prediction"]), classes)),
            np.concatenate((
                pred_column,
                prob_column
            )),
            f"LR Predictions with class probabilities (split test, evaluation set, overall score: {score})",
            mode
        )

    y_pred = model.predict(x_test)
    if isinstance(model, PLSR):
        y_pred = np.int8(np.abs(np.around(y_pred, decimals=0)))

    print_confusion_matrix(
        y1=y_test,
        y2=y_pred,
        title="Confusion matrix based on evaluation set",
        mode=mode
    )


def print_confusion_matrix(y1, y2, title: str, mode: GraphMode = GraphMode.GRAPHIC):
    '''Function to simplify the plotting of confusion matrices'''

    # Return early if there's nothing to print
    if mode is GraphMode.NONE:
        return

    cm = confusion_matrix(y1, y2)

    # Get unique class labels from y_true
    class_labels = sorted(set(y2))
    
    # Create the report
    report = classification_report(y1, y2, digits=2, output_dict=True)

    if mode is GraphMode.TEXT:
        print(f"\n{title}\n{cm}\n")

    elif mode is GraphMode.GRAPHIC:
        # Plot the confusion matrix using seaborn with custom colormap (Blues)
        sns.heatmap(
            cm,
            annot=True,
            fmt='d',
            cmap='Blues',
            xticklabels=class_labels,
            yticklabels=class_labels,
            cbar=False,
            vmin=0,
            vmax=cm.max()
        )

        plt.xlabel('Predicted')
        plt.ylabel('True')
        plt.title(title)
        plt.show()

    # Plot the classification report
    cr = pd.DataFrame(report).transpose()

    # Create a Plotly table
    print_table(
        ['substance'] + list(cr.columns),
        [cr.index, cr['precision'], cr['recall'], cr['f1-score'], cr['support']],
        "Classification Report",
        mode 
    )



'''This module contains the test cases for the base module.'''
import unittest

import os

import numpy as np

from chemfusekit.lldf import LLDFSettings, LLDF, Table
from chemfusekit.__base import BaseDataModel
from chemfusekit.lda import LDASettings, LDA


class TestBase(unittest.TestCase):
    def test_import_export(self):
        '''Test case for table import and export.'''

        # Import and fuse data from tables
        lldf_settings = LLDFSettings()
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        tables = [table1, table2]
        lldf = LLDF(lldf_settings, tables)
        lldf.lldf()

        # Export the fused dataset to file
        lldf.export_data("export_test.xlsx")

        # Import the fused dataset from file
        imported_data = BaseDataModel.load_from_file("export_test.xlsx", "Sheet1")

        # Assert the equality between the fused data and the export/import data
        # Compare the DataFrames using numpy.allclose()
        tolerance = 1e-6
        self.assertTrue(
            # Compare the DataFrames using numpy.allclose(): true if they match within tolerance
            np.allclose(lldf.fused_data.x_data.values, imported_data.x_data.values, atol=tolerance)
        )
        self.assertTrue(
            # The comparison between ndarrays returns an array of booleans, collapsed by "all()"
            (lldf.fused_data.y == imported_data.y).all()
        )

        # Second phase: re-export and re-import from BaseDataModel
        imported_data.export_to_file("export_test_2.xlsx")
        reimported_data = BaseDataModel.load_from_file("export_test_2.xlsx")

        # Assert the equality between the re-exported data and the re-reimported data
        self.assertTrue(np.allclose(lldf.fused_data.x_data.values, imported_data.x_data.values, atol=tolerance))
        self.assertTrue((lldf.fused_data.y == reimported_data.y).all())

        # Clean up
        os.remove("export_test.xlsx")
        os.remove("export_test_2.xlsx")

    def test_model_import(self):
        '''Integration test for model dumping and reloading.'''
        # Let's start by creating and training an LDA model
        lldf_settings = LLDFSettings()
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        tables = [table1, table2]
        lldf = LLDF(lldf_settings, tables)
        lldf.lldf()
        lda_settings = LDASettings()
        lda = LDA(lda_settings, lldf.fused_data)
        lda.lda()

        # Dump the model to file
        lda.export_model("modelfile.sklearn")

        # Reload the model
        lda2 = LDA(lda_settings, lldf.fused_data)
        lda2.import_model("modelfile.sklearn")

        # Check whether the imported model is the same as the exported model
        self.assertEqual(lda.model.get_params(), lda2.model.get_params())

        # Clean up
        os.remove("modelfile.sklearn")

    def test_from_file(self):
        '''Test case for classifier import from file'''
        '''Integration test for model dumping and reloading.'''
        # Let's start by creating and training an LDA model
        lldf_settings = LLDFSettings()
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        tables = [table1, table2]
        lldf = LLDF(lldf_settings, tables)
        lldf.lldf()
        lda_settings = LDASettings()
        lda = LDA(lda_settings, lldf.fused_data)
        lda.lda()

        # Dump the model to file
        lda.export_model("modelfile.sklearn")

        lda2 = LDA.from_file(lda_settings, "modelfile.sklearn")

        # Check whether the imported model is the same as the exported model
        self.assertEqual(lda.model.get_params(), lda2.model.get_params())

        # Clean up
        os.remove("modelfile.sklearn")

if __name__ == '__main__':
    unittest.main()



'''Linear Discriminant Analysis module'''
from copy import copy
from typing import Optional

import joblib
import numpy as np
import pandas as pd

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LD

from chemfusekit.lldf import LLDFDataModel
from chemfusekit.__utils import graph_output, run_split_test
from chemfusekit.__utils import print_confusion_matrix, print_table, GraphMode
from .__base import BaseDataModel, BaseClassifier, BaseSettings
from .pca import PCADataModel


class LDASettings(BaseSettings):
    '''Holds the settings for the LDA object.'''
    def __init__(self, components: int = 3, output: GraphMode = GraphMode.NONE, test_split: bool = False):
        super().__init__(output, test_split)
        if components <= 2:
            raise ValueError("Invalid component number: must be a > 1 integer.")
        self.components = components


class LDA(BaseClassifier):
    '''Class to store the data, methods and artifacts for Linear Discriminant Analysis'''
    def __init__(self, settings: LDASettings, data: BaseDataModel):
        super().__init__(settings, data)
        self.settings = settings
        self.data = data
        # Self-detect components if the data is from PCA
        if isinstance(data, PCADataModel):
            self.settings.components = data.components - 1

    def lda(self):
        '''Performs Linear Discriminant Analysis'''

        lda = LD(n_components=self.settings.components) # N-1 where N are the classes
        scores_lda = lda.fit(self.data.x_data, self.data.y).transform(self.data.x_data)
        pred = lda.predict(self.data.x_data)

        print_table(
            [f"LV{i+1}" for i in range(scores_lda.shape[1])],
            list(zip(*scores_lda)),
            "LDA scores",
            self.settings.output
        )

        print_table(
            ["Class", "Prior Probability"],
            [list(range(len(lda.priors_))), lda.priors_],
            "LDA Priors",
            self.settings.output
        )
        means_values = [[f"Feature {i+1}" for i in range(lda.means_.shape[1])]] + [list(lda.means_[i, :]) for i in range(lda.means_.shape[0])]
        
        print_table(
            ["Feature"] + [f"Class {i+1}" for i in range(lda.means_.shape[0])], 
            means_values,
            "LDA Class Means",
            self.settings.output
        )
        print_table(
            ["Component"] + [f"Feature {i+1}" for i in range(lda.coef_.shape[0])],
            [[f"Component {i+1}" for i in range(min(self.settings.components, lda.coef_.shape[1]))]] + [list(lda.coef_[:self.settings.components, i]) for i in range(lda.coef_.shape[0])],
            "LDA Coefficients",
            self.settings.output
        )
        classes, count = np.unique(pred, return_counts=True)
        print_table(
            ["Class", "Count"],
            [classes, count],
            "Prediction Counts",
            self.settings.output
        )

        pred = lda.predict(self.data.x_data)
        print_confusion_matrix(
            y1=self.data.y,
            y2=pred,
            title="LDA Training Confusion Matrix",
            mode=self.settings.output
        )

        lv_cols = [f'LV{i+1}' for i in range(self.settings.components)]
        scores = pd.DataFrame(data=scores_lda, columns=lv_cols)     # latent variables
        scores.index = self.data.x_data.index
        y_dataframe = pd.DataFrame(self.data.y, columns=['Substance'])

        scores = pd.concat([scores, y_dataframe], axis = 1)

        # Store the trained model
        self.model = lda

        # Show graphs if required by the user
        graph_output(
            scores,
            self.model,
            "Linear Discriminant Analysis",
            mode=self.settings.output
        )

        # Run split tests if required by the user
        if self.settings.test_split:
            run_split_test(
                scores.drop('Substance', axis=1).values,
                self.data.y,
                LD(n_components=self.settings.components),
                mode=self.settings.output
            )

    def import_model(self, import_path: str):
        model_backup = copy(self.model)
        super().import_model(import_path)
        if not isinstance(self.model, LD):
            self.model = model_backup
            raise ImportError("The file you tried to import is not a LinearDiscriminantAnalysis classifier.")
        self.settings.components = self.model.n_components



'''Performs low-level data fusion on input arrays, outputs the results'''
from typing import Optional, List

import numpy as np
import pandas as pd
from scipy.signal import savgol_filter

import matplotlib
import matplotlib.pyplot as plt

from .__base import GraphMode, BaseDataModel


class Table:
    '''Holds the path, preprocessing choice and sheet name for a single Excel table.'''
    def __init__(self, file_path: str, sheet_name: str, preprocessing: str, class_column: str = 'Substance',
                 index_column: str | None = None):
        self.file_path = file_path
        self.sheet_name = sheet_name
        self.preprocessing = preprocessing
        self.class_column = class_column
        self.index_column = index_column


class LLDFDataModel(BaseDataModel):
    '''Models the output data from the LLDF operation'''
    def __init__(self, x_data: pd.DataFrame, x_train: pd.DataFrame, y: np.ndarray):
        super().__init__(x_data, x_train, y)


class LLDFSettings:
    '''Holds the settings for the LLDF object.'''
    def __init__(self, output: GraphMode = GraphMode.NONE):
        self.output = output


def _snv(input_data: np.ndarray):
    '''Applies normalization to an input array'''
    # Define a new array and populate it with the corrected data
    output_data = np.zeros_like(input_data)
    for i in range(input_data.shape[0]):

        # Apply correction
        output_data[i, :] = (
            (input_data[i, :] - np.mean(input_data[i,:])) / np.std(input_data[i, :])
        )

    return output_data


class LLDF:
    '''Holds together all the data, methods and artifacts of the LLDF operation'''
    def __init__(self, settings: LLDFSettings, tables: List[Table]):
        self.settings = settings
        self.tables = tables
        self.fused_data: Optional[LLDFDataModel] = None

    def lldf(self):
        '''Performs low-level data fusion'''
        x_vector = []
        for table in self.tables:
            try:
                # Autodetect the format based on the file extension
                if table.file_path.endswith('.xlsx'):
                    table_data = pd.read_excel(
                        table.file_path,
                        sheet_name=table.sheet_name,
                        index_col=0,
                        header=0
                    )
                elif table.file_path.endswith('.csv'):
                    table_data = pd.read_csv(
                        table.file_path,
                        index_col=0,
                        header=0
                    )
                elif table.file_path.endswith('.json'):
                    table_data = pd.read_json(
                        table.file_path,
                        orient='table'  # or other orientations based on your json format
                    )
                else:
                    raise ValueError(f"Unsupported file format: {table.file_path}")
            except Exception as exc:
                raise FileNotFoundError("Error opening the selected files.") from exc

            # select only numerical attributes
            if table.index_column is not None and table.index_column in table_data.columns:
                x = table_data.drop(table.index_column, axis=1)
            else:
                x = table_data.iloc[:, 1:]

            if table.class_column in table_data.columns:
                x = table_data.drop(table.class_column, axis=1)
            else:
                x = x.iloc[:, 1:]

            # It is necessary to convert the column names as string to select them
            x.columns = x.columns.astype(str)     # to make the colnames as text

            # Preprocessing
            if table.preprocessing == 'snv':
                # Compute the SNV on spectra
                preprocessed_x = _snv(x.values)
            elif table.preprocessing == 'savgol':
                # Preprocessing with Savitzki-Golay
                # smoothing, defining the window, the order and the use of derivatives
                preprocessed_x = savgol_filter(x, 7, polyorder=2, deriv=0)
            elif table.preprocessing == 'savgol+snv':
                # We can also combine the preprocessing strategies together:
                # Savitzki-Golay - smoothing + SNV
                preprocessed_x = _snv(savgol_filter(x, 7, polyorder=2, deriv=0))
            elif table.preprocessing == 'none':
                # Skip preprocessing
                preprocessed_x = x
            else:
                raise SyntaxError(
                    f"LLDF: this type of preprocessing does not exist ({table.preprocessing=})"
                )

            if self.settings.output is GraphMode.GRAPHIC:
                numbers_string = [str(col) for col in x.columns]

                # Replace commas with points and join the numbers with a space
                try:
                    wl = np.array(list(map(lambda z: float(z.replace(',', '.')), numbers_string)))
                except ValueError:
                    wl = numbers_string

                if table.preprocessing != 'none':
                    # Let's plot the different datasets we preprocessed
                    fig, (ax1, ax2) = plt.subplots(2, figsize=(15, 15))
                    ax1.plot(wl, x.T)
                    ax1.set_title(f'Original data')
                    ax2.plot(wl, preprocessed_x.T)
                    ax2.set_title(f'Processed table with {table.preprocessing}')
                    if table.file_path.endswith('.xlsx'):
                        fig.suptitle(f'Imported table: {table.sheet_name} from {table.file_path}')
                    else:
                        fig.suptitle(f'Imported table: {table.file_path}')
                else:
                    # Let's plot the different datasets we preprocessed
                    fig, ax1 = plt.subplots(1, figsize=(15, 15))
                    ax1.plot(wl, x.T)
                    if table.file_path.endswith('.xlsx'):
                        fig.suptitle(f'Imported table: {table.sheet_name} from {table.file_path} (no preprocessing)')
                    else:
                        fig.suptitle(f'Imported table: {table.file_path} (no preprocessing)')

            # Create a new DataFrame with the processed numerical attributes
            processed_dataframe_x = pd.DataFrame(
                preprocessed_x,
                columns=x.columns
            )

            # Reset the index of the dataframe
            processed_dataframe_x = processed_dataframe_x.reset_index(drop=True)

            x_vector.append(processed_dataframe_x)

        try:
            table_data = pd.read_excel(
                self.tables[0].file_path,
                sheet_name=self.tables[0].sheet_name,
                index_col=0,
                header=0
            )
        except Exception as exc:
            raise FileNotFoundError("Error opening the selected files.") from exc

        y = table_data.loc[:, self.tables[0].class_column].values
        y_dataframe = pd.DataFrame(y, columns=['Substance'])

        # Fused dataset
        x_data = pd.concat(
            x_vector,
            axis=1
        )

        # X training set
        x_train = pd.concat(
            [y_dataframe, x_data],
            axis=1
        )

        self.fused_data = LLDFDataModel(x_data, x_train, y)

    def export_data(self, export_path: str, sheet_name: str = 'Sheet1'):
        '''Exports the data fusion artifacts to a file'''
        if self.fused_data is None:
            raise RuntimeError("Cannot export data before data fusion.")

        self.fused_data.export_to_file(export_path=export_path, sheet_name=sheet_name)



'''This module contains the test cases for the SVM module.'''
import unittest
from chemfusekit.svm import SVMSettings, SVM, GraphMode
from chemfusekit.lldf import LLDFSettings, LLDF, Table

class TestSVM(unittest.TestCase):
    '''Test suite for the LDA module.'''

    def test_svm_settings(self):
        '''Test case against settings errors.'''
        # Test against null type
        with self.assertRaises(TypeError):
            SVMSettings(None, GraphMode.NONE, False)
        # Test against null output selector
        with self.assertRaises(TypeError):
            SVMSettings('linear', None, False)
        # Test against null test_split selector
        with self.assertRaises(TypeError):
            SVMSettings('linear', GraphMode.NONE, None)
        # Test against non-existent kernels
        with self.assertRaises(ValueError):
            SVMSettings('non-existent', GraphMode.NONE, False)

        # Check if split tests with no output cause warnings:
        with self.assertRaises(Warning):
            SVMSettings(output=GraphMode.NONE, test_split=True)

        # Now call with proper values:
        SVMSettings(kernel='gaussian', output=GraphMode.GRAPHIC, test_split=False)

    def test_svm_constructor(self):
        '''Test case against constructor parameter issues.'''

        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        svm_settings = SVMSettings()

        # First, construct the object with null model:
        with self.assertRaises(TypeError):
            SVM(svm_settings, None)

        # Then, construct the object with null settings:
        with self.assertRaises(TypeError):
            SVM(None, lldf.fused_data)

        # Now, with both null:
        with self.assertRaises(TypeError):
            SVM(None, None)

        # Finally, with proper values:
        svm = SVM(svm_settings, lldf.fused_data)
    
    def test_svm(self):
        '''Integration test case.'''

        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Create an SVM object and train it, with no output
        svm_settings = SVMSettings(output=GraphMode.NONE)
        svm = SVM(svm_settings, lldf.fused_data)
        svm.svm()

        # Create an SVM object and train it, with graphical output
        svm_settings = SVMSettings(output=GraphMode.GRAPHIC)
        svm = SVM(svm_settings, lldf.fused_data)
        svm.svm()

        # Create an SVM object and train it, with text output
        svm_settings = SVMSettings(output=GraphMode.TEXT)
        svm = SVM(svm_settings, lldf.fused_data)
        svm.svm()

    def test_svm_predict(self):
        '''Test case against prediction parameter issues.'''

        # Perform preliminary data fusion
        lldf_settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path="tests/qepas.xlsx",
            sheet_name="Sheet1",
            preprocessing="snv"
        )
        table2 = Table(
            file_path="tests/rt.xlsx",
            sheet_name="Sheet1",
            preprocessing="none"
        )
        lldf = LLDF(lldf_settings, [table1, table2])
        lldf.lldf()

        # Create an SVM object without training it
        svm_settings = SVMSettings()
        svm = SVM(svm_settings, lldf.fused_data)

        # Pick a random sample for prediction
        x_data_sample = lldf.fused_data.x_train.iloc[119] # should be DMMP
        x_data_sample = x_data_sample.iloc[1:].to_frame().transpose()

        # Run prediction with untrained model (should throw exception)
        with self.assertRaises(RuntimeError):
            svm.predict(x_data_sample)

        # Train the SVM object
        svm.svm()

        # Run prediction with empty data (should throw exception)
        with self.assertRaises(TypeError):
            svm.predict(None)

        # Run prediction with trained model and non-null data
        svm.predict(x_data_sample)



'''This module contains the test cases for the LLDF module.'''
import unittest
from chemfusekit.lldf import LLDFSettings, LLDF, GraphMode, Table

class TestLLDF(unittest.TestCase):
    '''Test suite for the LLDF module.'''

    def test_file_loading(self):
        '''Test case against file loading errors.'''
        # load a non-existent file on purpose
        settings = LLDFSettings(
            output=GraphMode.NONE
        )

        table1 = Table(
            file_path='tests/nonexistent.xlsx',
            sheet_name='Sheet1',
            preprocessing='none'
        )

        files = [table1]
        lldf = LLDF(settings=settings, tables=files)
        self.assertRaises(FileNotFoundError, lldf.lldf)

    def test_preprocessing_techniques(self):
        '''Test case against wrong preprocessing user input.'''
        with self.assertRaises(SyntaxError):
            settings = LLDFSettings(
                output=GraphMode.NONE
            )

            table1 = Table(
                file_path='tests/qepas.xlsx',
                sheet_name='Sheet1',
                preprocessing='qpl'
            )

            lldf = LLDF(settings, [table1])
            lldf.lldf()

        # Now a correct value:
        settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path='tests/qepas.xlsx',
            sheet_name='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(settings, [table1])
        lldf.lldf()
    
    def test_export(self):
        '''Test case against wrong export settings.'''
        settings = LLDFSettings(output=GraphMode.NONE)
        table1 = Table(
            file_path='tests/qepas.xlsx',
            sheet_name='Sheet1',
            preprocessing='snv'
        )
        lldf = LLDF(settings, [table1])
        
        # Try exporting data before data fusion
        with self.assertRaises(RuntimeError):
            lldf.export_data('path')

        # Perform data fusion
        lldf.lldf()

        # Try exporting data to an invalid path
        with self.assertRaises(ValueError):
            lldf.export_data('$£=0\//|') 

if __name__ == '__main__':
    unittest.main()



'''Partial Least Squares Discriminant Analysis module.'''
from copy import copy
from typing import Optional

import pandas as pd
import numpy as np

import plotly.express as px

from sklearn.cross_decomposition import PLSRegression as PLSR

from chemfusekit.lldf import LLDFDataModel
from chemfusekit.__utils import GraphMode, print_table, print_confusion_matrix, run_split_test
from .__base import BaseSettings, BaseDataModel, BaseClassifier


class PLSDASettings(BaseSettings):
    '''Holds the settings for the PLSDA object.'''
    def __init__(self, n_components: int = 3, output: GraphMode = GraphMode.NONE, test_split: bool = False):
        super().__init__(output, test_split)
        if n_components < 1:
            raise ValueError("Invalid n_components number: should be a positive integer.")
        self.n_components = n_components


class PLSDA(BaseClassifier):
    '''
    Class to store the data, methods and artifacts for Partial Least Squares
    Discriminant Analysis
    '''
    def __init__(self, settings: PLSDASettings, data: BaseDataModel):
        super().__init__(settings, data)

    def plsda(self):
        '''Performs Partial Least Squares Discriminant Analysis'''
        x = self.data.x_data
        y = self.data.x_train.Substance.astype('category').cat.codes

        regr_pls = PLSR(n_components=self.settings.n_components)
        regr_pls.fit_transform(x, y)

        # Save the model
        self.model = copy(regr_pls)

        # Re-create the model
        regr_pls_2 = PLSR(n_components=self.settings.n_components)
        regr_pls_2.fit_transform(x, y)

        # Scores and loadings
        lv_cols = [f"LV{i + 1}" for i in range(self.settings.n_components)]
        scores = pd.DataFrame(regr_pls_2.x_scores_, columns=lv_cols)
        scores.index = self.data.x_train.index
        y = self.data.x_train.Substance
        scores = pd.concat([scores, y], axis=1)

        print_table(
            scores.columns,
            scores.transpose(),
            "PLSDA scores by substance",
            self.settings.output
        )

        # Loadings
        loadings = pd.DataFrame(regr_pls.x_loadings_, columns=lv_cols)
        loadings["Attributes"] = self.data.x_train.iloc[:, 1:].columns
        print_table(
            loadings.columns,
            loadings.transpose(),
            "PLSDA Loadings",
            self.settings.output
        )

        if self.settings.output is GraphMode.GRAPHIC:
            # Scores plot
            fig = px.scatter(scores, x="LV1", y="LV2", color="Substance", hover_data=['Substance'])
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_layout(
                height=600,
                width=800,
                title_text='Scores Plot')
            fig.show()

            # Plot 3D scores
            fig = px.scatter_3d(scores, x='LV1', y='LV2', z='LV3',
                                color='Substance', hover_data=['Substance'],
                                hover_name=scores.index)
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_layout(
                title_text='Scores 3D Plot colored by Class using the raw data')
            fig.show()

            # Loadings plot
            fig = px.scatter(loadings, x="LV1", y="LV2",text="Attributes")
            fig.update_xaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_yaxes(zeroline=True, zerolinewidth=1, zerolinecolor='Black')
            fig.update_traces(textposition='top center')
            fig.update_layout(
                height=600,
                width=800,
                title_text='Loadings Plot')
            fig.show()

            # Predictions
            pred = regr_pls.predict(x)
            pred = np.int8(np.round(pred, decimals=0))
            classes = np.unique(y)
            for i in range(len(np.unique(pred)) - len(classes)):
                classes = np.append(classes, f'Unknown substance {i}')
            predicted_substances = [classes[p-3] for p in pred]
            print_table(
                ["Sample number", "True", "Predicted"],
                np.stack((
                    [f"{i+1}" for i in range(len(pred))],
                    y,
                    predicted_substances
                )),
                "True and predicted substances",
                self.settings.output
            )

            # Print confusion matrix
            y = self.data.x_train.Substance.astype('category').cat.codes
            print_confusion_matrix(
                y,
                pred,
                "Confusion Matrix based on training set",
                self.settings.output
            )

        if self.settings.output and self.settings.test_split:
            x = self.data.x_data
            y = self.data.x_train.Substance.astype('category').cat.codes
            run_split_test(x, y, PLSR(self.settings.n_components), mode=self.settings.output)

    def import_model(self, import_path: str):
        model_backup = copy(self.model)
        super().import_model(import_path)
        if not isinstance(self.model, PLSR):
            self.model = model_backup
            raise ImportError("The file you tried to import is not a PLSRegression classifier.")
        self.settings.n_components = self.model.n_components
